/* BdfServer - Copyright (c) 2015-2023 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.bdfserver.tools.ui.components;

import fr.exemole.bdfserver.api.BdfServerConstants;
import fr.exemole.bdfserver.api.ui.FieldUi;
import fr.exemole.bdfserver.tools.ui.options.InputUiOptionBuilder;
import fr.exemole.bdfserver.tools.ui.options.UiOptionsUtils;
import fr.exemole.bdfserver.tools.ui.options.ValueUiOptionBuilder;
import net.fichotheque.corpus.metadata.FieldKey;
import net.mapeadores.util.annotation.Nullable;
import net.mapeadores.util.attr.AttributeChangeBuilder;
import net.mapeadores.util.attr.Attributes;
import fr.exemole.bdfserver.api.ui.options.InputUiOption;
import fr.exemole.bdfserver.api.ui.options.ValueUiOption;


/**
 *
 * @author Vincent Calame
 */
public class FieldUiBuilder {

    private final FieldKey fieldKey;
    private final AttributeChangeBuilder attributeChangeBuilder = new AttributeChangeBuilder();
    private final Attributes previousAttributes;
    private final InputUiOptionBuilder inputUiOptionBuilder;
    private final ValueUiOptionBuilder valueUiOptionBuilder;
    private String status = BdfServerConstants.DEFAULT_STATUS;

    public FieldUiBuilder(FieldKey fieldKey) {
        this(fieldKey, null);
    }

    public FieldUiBuilder(FieldKey fieldKey, @Nullable Attributes previousAttributes) {
        if (fieldKey == null) {
            throw new IllegalArgumentException("corpusField is null");
        }
        this.fieldKey = fieldKey;
        if (FieldUi.hasInputUiOption(fieldKey)) {
            this.inputUiOptionBuilder = new InputUiOptionBuilder(UiOptionsUtils.getDefaultInputUiOption(fieldKey));
        } else {
            this.inputUiOptionBuilder = null;
        }
        if (FieldUi.hasValueUiOption(fieldKey)) {
            this.valueUiOptionBuilder = new ValueUiOptionBuilder();
        } else {
            this.valueUiOptionBuilder = null;
        }
        this.previousAttributes = previousAttributes;
    }

    public FieldUiBuilder(FieldUi fieldUi) {
        this.fieldKey = fieldUi.getFieldKey();
        this.status = fieldUi.getStatus();
        InputUiOption inputUiOption = fieldUi.getInputUiOption();
        if (inputUiOption != null) {
            this.inputUiOptionBuilder = new InputUiOptionBuilder(inputUiOption);
        } else {
            this.inputUiOptionBuilder = null;
        }
        ValueUiOption valueUiOption = fieldUi.getValueUiOption();
        if (valueUiOption != null) {
            this.valueUiOptionBuilder = new ValueUiOptionBuilder(valueUiOption);
        } else {
            this.valueUiOptionBuilder = null;
        }
        this.previousAttributes = fieldUi.getAttributes();
    }

    public FieldUiBuilder setStatus(String status) {
        if (((status != null) && (FieldUi.isModifiableStatus(fieldKey)))) {
            this.status = BdfServerConstants.checkValidStatus(status);
        }
        return this;
    }

    /**
     * Null si non approprié.
     *
     * @return
     */
    public InputUiOptionBuilder getInputUiOptionBuilder() {
        return inputUiOptionBuilder;
    }

    public ValueUiOptionBuilder getValueUiOptionBuilder() {
        return valueUiOptionBuilder;
    }

    public AttributeChangeBuilder getAttributeChangeBuilder() {
        return attributeChangeBuilder;
    }

    public FieldUi toFieldUi() {
        InputUiOption inputUiOption = null;
        ValueUiOption valueUiOption = null;
        if (inputUiOptionBuilder != null) {
            inputUiOption = inputUiOptionBuilder.toInputUiOption();
        }
        if (valueUiOptionBuilder != null) {
            valueUiOption = valueUiOptionBuilder.toValueUiOption();
        }
        return new InternalFieldUi(fieldKey, status, inputUiOption, valueUiOption, attributeChangeBuilder.toAttributes(previousAttributes));
    }

    public static FieldUiBuilder init(FieldKey fieldKey) {
        return new FieldUiBuilder(fieldKey, null);
    }

    public static FieldUiBuilder init(FieldKey fieldKey, @Nullable Attributes initAttributes) {
        return new FieldUiBuilder(fieldKey, initAttributes);
    }

    public static FieldUiBuilder derive(FieldUi fieldUi) {
        return new FieldUiBuilder(fieldUi);
    }


    private static class InternalFieldUi implements FieldUi {

        private final FieldKey fieldKey;
        private final String status;
        private final InputUiOption inputUiOption;
        private final ValueUiOption valueUiOption;
        private final Attributes attributes;

        private InternalFieldUi(FieldKey fieldKey, String status, InputUiOption inputUiOption, ValueUiOption valueUiOption, Attributes attributes) {
            this.fieldKey = fieldKey;
            this.status = status;
            this.inputUiOption = inputUiOption;
            this.valueUiOption = valueUiOption;
            this.attributes = attributes;
        }

        @Override
        public FieldKey getFieldKey() {
            return fieldKey;
        }

        @Override
        public String getStatus() {
            return status;
        }

        @Override
        public InputUiOption getInputUiOption() {
            return inputUiOption;
        }

        @Override
        public ValueUiOption getValueUiOption() {
            return valueUiOption;
        }

        @Override
        public String getName() {
            return fieldKey.getKeyString();
        }

        @Override
        public Attributes getAttributes() {
            return attributes;
        }

    }

}
